"""Smartdoor system including NFC card detecting, key locking/unlocking, turning LED on/off, etc.

This system is designed to be used with Raspberry Pi.

Some CLIs including main sequence is implemented here.
"""
from __future__ import annotations

import subprocess
from importlib.util import find_spec
from logging import config as log_config
from logging import getLogger
from pathlib import Path
from pprint import pformat

import rich_click as click

# Define logger before importing SmartDoor class to follow the same configuration
log_config.fileConfig(Path(__file__).parent / "logging.conf")
logger = getLogger("main")

try:
    import tomllib
except ImportError:
    import tomli as tomllib

from .smartdoor import SmartDoor

__version__ = "2.0.1"
__all__ = ["SmartDoor"]


@click.group(context_settings=dict(help_option_names=["-h", "--help"]))
@click.version_option(__version__, "-V", "--version")
def cli():
    """Smartdoor system CLI."""
    pass


@cli.command()
@click.option(
    "--locked/--unlocked", default=True, show_default=True, help="Set initial key status."
)
def start(locked: bool):
    """Start SmartDoor system.

    The infinite loop workflow is executed at foreground. The initial key status can be set by
    `--locked` or `--unlocked` option, by default `--locked`.

    If you want to stop this system, press Ctrl+C.
    """
    # Instantiate SmartDoor
    logger.info("start smartdoor system")
    door = SmartDoor()

    # Set initial key status
    door.locked = locked
    logger.info("Set initial key status to %s", "locked" if locked else "unlocked")

    # SmartDoor seaquence starts
    try:
        while True:
            tag = door.wait_for_touched()

            # If buttom is pushed
            if tag is None:
                door.led_button.blink(on_time=0.2, off_time=0.2)
                door.door_sequence(user="Button operator")
                door.led_button.on()
                continue

            # If NFC card is detected
            elif bool(tag):
                user = door.authenticate(tag)

                # If invalid user is detected
                if user is None:
                    door.warning_sequence()
                    continue
                else:
                    door.door_sequence(user=user)

            # If tag == False (KeyboardInterrupt)
            else:
                raise KeyboardInterrupt
    except KeyboardInterrupt:
        logger.info("Smartdoor system stopped by user")

    except Exception:
        logger.exception("unexpected error occurred")
        door.error_sequence()

    finally:
        door.close()


@cli.command()
@click.option("--debug", "-d", is_flag=True, help="show debug log")
def logs(debug: bool):
    """Show logs of SmartDoor system.

    logs are stored in `~/smartdoor.log` or `~/smartdoor_debug.log`. If you want to show debug log,
    use `--debug` option.
    """
    log = Path().home() / "smartdoor.log" if not debug else Path().home() / "smartdoor_debug.log"

    if log.exists():
        click.echo(log.read_text())
    else:
        click.echo("log file not found.")


@cli.command()
@click.option("--show", is_flag=True, help="show current configuration parameters")
@click.option(
    "--generate", is_flag=True, help="generate default config file as ~/.config/smartdoor.toml"
)
def config(show: bool, generate: bool):
    """Configuration tool for SmartDoor system.

    If you want to configure smartdoor system, edit `~/.config/smartdoor.toml` directly, after
    generating default config file by `--generate` option.
    """
    # Load default configuration file
    default_config_path = Path(__file__).parent / "default_config.toml"
    with default_config_path.open("rb") as file:
        config = tomllib.load(file)

    # Load user-specific configuration file if exists
    config_path = Path.home() / ".config" / "smartdoor.toml"
    if config_path.exists():
        with config_path.open("rb") as file:
            config.update(tomllib.load(file))

    # If config file not found and `--generate` option is specified, generate default config file
    if generate and not config_path.exists():
        config_path.parent.mkdir(parents=True, exist_ok=True)
        with config_path.open("w") as file:
            file.write(default_config_path.read_text())
        click.echo(f"generated default config file as {config_path}")

    # Show configuation
    elif show:
        click.echo(pformat(config))
    else:
        click.echo(click.get_current_context().get_help())


@cli.command()
@click.option("--register", is_flag=True, help="register service to systemd")
@click.option("--unregister", is_flag=True, help="unregister service from systemd")
@click.option("--start", is_flag=True, help="start service")
@click.option("--stop", is_flag=True, help="stop service")
@click.option("--restart", is_flag=True, help="restart service")
@click.option("--status", is_flag=True, help="show service status")
def service(register: bool, unregister: bool, start: bool, stop: bool, restart: bool, status: bool):
    """Systemd Service tool for SmartDoor system.

    If you want to register/unregister smartdoor system to/from systemd, use `--register` or
    `--unregister` option. If choosing `--register` option, smartdoor system will be ready for start
    automatically.

    You can also start/stop/restart service by `--start`, `--stop`, `--restart` option. If you want
    to show service status, use `--status` option.

    If `pigpio` is installed, launching `pigpio` daemon is also registered to systemd automatically.
    """
    if register:
        # Register pigpio daemon to systemd
        if find_spec("pigpio") is not None:
            try:
                service_file = Path(__file__).parent / "pigpio.service"
                subprocess.run(
                    ["sudo", "systemctl", "link", str(service_file)],
                    check=True,
                )
                subprocess.run(["sudo", "systemctl", "daemon-reload"], check=True)
                subprocess.run(["sudo", "systemctl", "enable", "pigpio.service"], check=True)
                click.echo("registered pigpio daemon to systemd")
            except subprocess.CalledProcessError as e:
                click.echo(e)
                click.echo("failed to register pigpio daemon to systemd")

        # Register smartdoor service to systemd
        try:
            service_file = Path(__file__).parent / "smartdoor.service"
            subprocess.run(
                ["sudo", "systemctl", "link", str(service_file)],
                check=True,
            )
            subprocess.run(["sudo", "systemctl", "daemon-reload"], check=True)
            subprocess.run(["sudo", "systemctl", "enable", "smartdoor.service"], check=True)
            click.echo("registered service to systemd")
        except subprocess.CalledProcessError as e:
            click.echo(e)
            click.echo("failed to register service to systemd")

    elif unregister:
        # Unregister smartdoor service from systemd
        try:
            subprocess.run(["sudo", "systemctl", "stop", "smartdoor.service"], check=True)
            subprocess.run(["sudo", "systemctl", "disable", "smartdoor.service"], check=True)
            subprocess.run(["sudo", "systemctl", "daemon-reload"], check=True)
            click.echo("unregistered service from systemd")
        except subprocess.CalledProcessError as e:
            click.echo(e)
            click.echo("failed to unregister service from systemd")

        # Unregister pigpio daemon from systemd
        if find_spec("pigpio") is not None:
            try:
                subprocess.run(["sudo", "systemctl", "stop", "pigpio.service"], check=True)
                subprocess.run(["sudo", "systemctl", "disable", "pigpio.service"], check=True)
                subprocess.run(["sudo", "systemctl", "daemon-reload"], check=True)
                click.echo("unregistered pigpio daemon from systemd")
            except subprocess.CalledProcessError as e:
                click.echo(e)
                click.echo("failed to unregister pigpio daemon from systemd")

    elif start:
        # Start pigpio daemon
        if find_spec("pigpio") is not None:
            try:
                subprocess.run(["sudo", "systemctl", "start", "pigpio.service"], check=True)
                click.echo("started pigpio daemon")
            except subprocess.CalledProcessError as e:
                click.echo(e)
                click.echo("failed to start pigpio daemon")

        # Start smartdoor service
        try:
            subprocess.run(["sudo", "systemctl", "start", "smartdoor.service"], check=True)
            click.echo("started service")
        except subprocess.CalledProcessError as e:
            click.echo(e)
            click.echo("failed to start service")

    elif stop:
        # Stop smartdoor service
        try:
            subprocess.run(["sudo", "systemctl", "stop", "smartdoor.service"], check=True)
            click.echo("stopped service")
        except subprocess.CalledProcessError as e:
            click.echo(e)
            click.echo("failed to stop service")

        # Stop pigpio daemon
        if find_spec("pigpio") is not None:
            try:
                subprocess.run(["sudo", "systemctl", "stop", "pigpio.service"], check=True)
                click.echo("stopped pigpio daemon")
            except subprocess.CalledProcessError as e:
                click.echo(e)
                click.echo("failed to stop pigpio daemon")

    elif restart:
        try:
            subprocess.run(["sudo", "systemctl", "restart", "smartdoor.service"], check=True)
            click.echo("restarted service")
        except subprocess.CalledProcessError as e:
            click.echo(e)
            click.echo("failed to restart service")

    elif status:
        try:
            subprocess.run(["sudo", "systemctl", "status", "smartdoor.service"], check=True)
        except subprocess.CalledProcessError as e:
            click.echo(e)
            click.echo("failed to show service status")

    else:
        click.echo(click.get_current_context().get_help())


if __name__ == "__main__":
    cli()



"""This module provides IDm authentication functions communicating with database through web api."""
from __future__ import annotations

import sys
from logging import getLogger

from requests import Request, Session

module_logger = getLogger(__name__)


class AuthIDm:
    """Authentication for IDm using web api.

    Parameters
    ----------
    url
        URL address to post IDm data
    room
        name of room to authenticate, e.g. "423"
    timeout
        session timeout to connect to the database, by default 10 sec
    """

    logger = getLogger("main").getChild("AuthIDm")

    def __init__(self, url: str, room: str, timeout: float = 10) -> None:
        try:
            # get CSRF token in cookies
            session = Session()
            session.get(url, timeout=timeout, verify=False)

            # instantiate request object
            headers = {
                "Content-type": "application/json",
                "X-CSRFToken": session.cookies["csrftoken"],
            }
            req = Request("POST", url, headers=headers)

        except Exception as e:
            self.logger.error(f"cannot establish the connection to {url}")
            self.logger.debug(f"{type(e)}: {e}")
            sys.exit(1)

        # save variables as properties
        self._session = session
        self._request = req
        self._room = room

    @property
    def session(self) -> Session:
        """Session object."""
        return self._session

    @property
    def request(self) -> Request:
        """Request object."""
        return self._request

    @property
    def room(self) -> str:
        """Room name."""
        return self._room

    def authenticate(self, idm: str, timeout: float = 5) -> str | None:
        """IDm authentication method.

        If the given idm is validated, returns user name which is registered in database.
        Otherwise returns None.

        Parameters
        ----------
        idm
            Hexadecimal IDm converted to string
        timeout
            post requesting timeout to the database, by default 5 sec

        Returns
        -------
        str | None
            username which is registered in database, if not, retuning None.
        """
        if not isinstance(idm, str):
            self.logger.error("idm must be string containing 16 hexadecimal digits.")
            return None

        self._request.json = {"idm": idm}

        # request preparation
        prepared_req = self.session.prepare_request(self._request)
        try:
            response = self._session.send(prepared_req, timeout=timeout, verify=False)
            data = response.json()

            if data["auth"] == "valid":
                if data[f"allow_{self._room}"]:
                    return data["name"]
            return None

        except Exception as e:
            self.logger.error(f"{type(e)}: {e}")
            return None

    def close(self):
        """Close session."""
        self.logger.debug("close session")
        self._session.close()


# for debug
if __name__ == "__main__":
    url = "http://localhost:8000/authenticate/"
    room = "423"
    auth = AuthIDm(url, room)

    print(auth.authenticate("013BDD2FEE1FC80D"))
    auth.close()



"""This module provides a main class of SmartDoor system."""
from __future__ import annotations

from binascii import hexlify
from collections import deque
from datetime import datetime
from logging import getLogger
from pathlib import Path
from time import sleep

import requests
from nfc import ContactlessFrontend
from nfc.tag import Tag
from urllib3 import disable_warnings
from urllib3.exceptions import InsecureRequestWarning

try:
    import tomllib
except ImportError:
    import tomli as tomllib

from .core import AuthIDm, SmartLock

# set invisible of https secure warning
disable_warnings(InsecureRequestWarning)

module_logger = getLogger(__name__)


class SmartDoor(SmartLock):
    """Smart Door system class.

    This class has some main fetures of door lock/unlock seaquence, reading/authenticate NFC's IDm,
    and post to IFTTT for notification.

    The user-specific configuration file (`~/.config/smartdoor.toml`) is loaded automatically.
    """

    # define class logger
    logger = getLogger("main").getChild("SmartDoor")

    def __init__(self) -> None:
        # Load default configuration file
        with open(Path(__file__).parent / "default_config.toml", "rb") as file:
            config = tomllib.load(file)

        # Load user-specific configuration file if exists
        config_path = Path.home() / ".config" / "smartdoor.toml"
        if config_path.exists():
            with config_path.open("rb") as file:
                config.update(tomllib.load(file))
                self.logger.debug(f"Loaded user-specific configuration file: {config_path}")

        # === Initialization ==================================================
        # IFTTT
        self._urls: dict[str, str] = {}
        self.urls = config["IFTTT_URLs"]
        self._post_queue: deque = deque()

        # IDm authentication class
        self._auth = AuthIDm(config["auth_url"], config["room"])

        # NFC reader
        self._clf = ContactlessFrontend("usb")

        # room name
        self._room = config["room"]

        # inheritance
        super().__init__(config["pins"])

    @property
    def urls(self) -> dict[str, str]:
        """URL map to post to IFTTT.

        key is event name, value is URL string.
        """
        return self._urls

    @urls.setter
    def urls(self, value):
        if not isinstance(value, dict):
            raise TypeError(f"Invalid type of urls: {type(value)}")

        for key, url in value.items():
            if not isinstance(key, str):
                raise TypeError(f"Invalid type of key: {type(key)}")
            if not isinstance(url, str):
                raise TypeError(f"Invalid type of url: {type(url)}")
            self._urls[key] = url

    @property
    def post_queue(self) -> deque:
        """Queue to store values to post them to IFTTT."""
        return self._post_queue

    @property
    def clf(self) -> ContactlessFrontend:
        """Nfcpy's Contactless Frontend class instance."""
        return self._clf

    @property
    def auth(self) -> AuthIDm:
        """Instance of AuthIDm class."""
        return self._auth

    def wait_for_touched(self) -> None | bool | Tag:
        """Wait for the NFC card to be touched.

        Returns
        -------
        None | bool | obj:`nfc.tag.Tag`
            If the button is pushed, returns None.
            If the KeyboadInterrupt is detected, returns False.
            otherwise, returns the instance of nfcpy's Tag class.
        """
        rdwr_options = {
            "targets": ["212F"],  # detect only Felica
            "on-connect": lambda tag: False,
            "iterations": 5,
            "interval": 0.5,
        }
        tag = self.clf.connect(rdwr=rdwr_options, terminate=lambda: self.button.is_pressed)

        return tag

    def authenticate(self, tag: Tag) -> str | None:
        """Authenticate the approved user.

        This method authenticates the user by the IDm of the NFC card.

        Parameters
        ----------
        tag
            the instance of nfcpy's Tag class

        Returns
        -------
        str | None
            If the user is approved, returns the user name.
        """
        # turn on both red and green LEDs
        if self.locked:
            self.led_green.on()
        else:
            self.led_red.on()

        # extract idm
        idm = hexlify(tag.idm).decode("utf-8")

        self.logger.debug(f"idm: {idm} is detected.")

        # authentication
        name = self._auth.authenticate(idm)

        return name

    def post_ifttt(self, user: str = "test", action: str = "LOCK") -> None:
        """Post the info to IFTTT.

        This method posts the info:``{"value1": data, "value2": user, "value3": action}``
        in a json format to IFTTT urls.
        The action means the smartdoor action like a "LOCK", "UNLOCK", etc.

        If the post is failed, the data is cached into a queue and try to post again.

        Parameters
        ----------
        user
            user name, by default "test"
        action
            smartdoor action like "LOCK", "UNLOCK", etc, by default "LOCK"
        """
        # get current datetime
        date_now = datetime.now()
        date_str = date_now.strftime(r"%Y/%m/%d %H:%M:%S")

        # cache post values into a queue
        values = {"value1": date_str, "value2": user, "value3": action}
        self.post_queue.append(values)

        # post to IFTTT in 3 seconds connect timeout and 7.5 seconds read timeout
        try:
            while self.post_queue:
                data = self.post_queue.popleft()

                for event, url in self.urls.items():
                    res = requests.post(url, json=data, timeout=(3.0, 7.5))

                    if res.status_code == 200:
                        self.logger.debug(f"IFTTT post is completed: {event}")
                    else:
                        self.logger.error(
                            f"IFTTT post is failed (code: {res.status_code}): {event}"
                        )
                        self._post_queue.appendleft(data)
                        break

            self.logger.info("IFTTT post is completed.")

        except Exception:
            self.logger.exception("IFTTT post is failed.")
            self._post_queue.appendleft(data)

    def door_sequence(self, user: str = "test") -> None:
        """Door sequence.

        If the door is locked, :obj:`.unlock` method is excuted.
        Otherwise, :obj:`.lock` method is excuted.
        After that, the info is posted to IFTTT by :obj:`.post_ifttt` method.

        Parameters
        ----------
        user
            user name controlling the door, by default "test"
        """
        if self.locked:
            self.unlock()
            action = "UNLOCK"
            self.logger.info(f"unlocked by {user}")
        else:
            self.lock()
            action = "LOCK"
            self.logger.info(f"locked by {user}")

        # post to IFTTT
        self.post_ifttt(user=user, action=action)

        # pause for 1 sec to avoid multiple execution
        sleep(1)

    def warning_sequence(self) -> None:
        """Warning sequence when an unauthorized user touched the reader."""
        # Blink red LED and sound buzzer
        self.led_green.off()
        self.led_red.blink(on_time=0.1, off_time=0.1)
        self.buzzer.beep(n=2, on_time=0.5, off_time=0.1)

        # log
        self.logger.info("unauthorized user touched the reader")

        # post to IFTTT
        self.post_ifttt(user="unauthorized user", action="INVALID TOUCH")

        # restore LED
        if self.locked:
            self.led_red.on()
        else:
            self.led_red.off()
            self.led_green.on()

    def error_sequence(self, duration: float = 2) -> None:
        """Error sequence when an program error occurred.

        The sequence is as follows:
        - Blink all LED
        - Sound buzzer for `duration` seconds
        - After that, turn off all LED and buzzer

        Parameters
        ----------
        duration
            duration of the error sequence, by default 2 sec
        """
        # Blink all LED
        self.led_red.blink(on_time=0.1, off_time=0.1)
        self.led_green.blink(on_time=0.1, off_time=0.1)

        # sound buzzer
        self.buzzer.on()
        sleep(duration)

        # Off all LED and buzzer
        self.buzzer.off()
        self.led_red.off()
        self.led_green.off()

    def close(self) -> None:
        """Close the smartdoor system."""
        try:
            self.clf.close()  # close nfc contactlessfrontend instance
            self._auth.close()  # close authentication session
            self.logger.info("smartdoor system is closed.")
        except Exception:
            self.logger.exception("failed to close smartdoor system.")



"""Module for SmartLock class."""
from __future__ import annotations

from logging import getLogger
from time import sleep

from gpiozero import LED, Button, Buzzer, Device, Servo

module_logger = getLogger(__name__)


class SmartLock:
    """This class is used to control raspberry Pi's GPIO devices (LEDs, Buzzer, servomotor, etc.).

    Parameters
    ----------
    pins
        pin assignment key-value map. A pin number must be GPIO pin not BCM one.

        The following keys must be included:

        - button: input signal from push button switch
        - LED_red : output signal to red LED
        - LED_green : output signal to green LED
        - LED_button : output signal to button switch LED
        - buzzer : output signal to buzzer
        - servo : output signal to servomotor

    Methods
    -------
    lock()
        excute lock seaquence
        (LED blinking -> buzzer beeping -> servomotor moving -> LED lighting)
    unlock()
        excute unlock seaquence
        (LED blinking -> buzzer beeping -> servomotor moving -> LED lighting)
    """

    # define class logger
    logger = getLogger("main").getChild("SmartLock")

    def __init__(self, pins: dict[str, int]) -> None:
        # set pin factory
        try:
            from gpiozero.pins.pigpio import PiGPIOFactory

            Device.pin_factory = PiGPIOFactory()
            self.logger.debug("using pigpio pin factory")

        except Exception:
            self.logger.debug("using defalut pin factory")

        # validate pin assignment
        self.pins = pins

        # Initialize GPIO devices
        # === LED ===
        self.led_red = LED(pins["LED_red"])
        self.led_green = LED(pins["LED_green"])
        self.led_button = LED(pins["LED_button"])
        self.led_button.on()

        # === Push button switch ===
        self.button = Button(pins["button"])

        # === Buzzer ===
        self.buzzer = Buzzer(pins["buzzer"])

        # === Servo ===
        self.servo = Servo(
            pins["servo"],
            initial_value=None,
            min_pulse_width=0.5e-3,
            max_pulse_width=2.4e-3,
        )

        # initialize locked property to avoid unexpected behavior
        self._locked = False

    @property
    def pins(self) -> dict[str, int]:
        """Pins asssignment map."""
        return self._pins

    @pins.setter
    def pins(self, maps):
        if not isinstance(maps, dict):
            raise TypeError("pins must be dict type.")

        # validate pin assignment
        self._check_pin_overlap(maps)
        self._pins = maps

    @property
    def locked(self) -> bool:
        """Key's status.

        If true, the key is locked now, otherwise unlocked

        Returns
        -------
        bool
            key's status (True: locked, False: unlocked)
        """
        return self._locked

    @locked.setter
    def locked(self, value):
        if not isinstance(value, bool):
            raise TypeError("locked must be a bool value")
        self._locked = value
        # update LED status
        if self._locked:
            self.led_red.on()
            self.led_green.off()
        else:
            self.led_red.off()
            self.led_green.on()

    def lock(self) -> None:
        """Excute Locking sequence.

        The detail of the sequence is as follows:

        1. Green LED off
        2. Red LED Blinking
        3. Buzzer beeping (2 times)
        4. Servomotor moving
        5. Red LED on
        """
        self.logger.debug("locking sequence started")

        # Green LED off
        self.led_green.off()

        # Red LED blinking
        self.led_red.blink(on_time=0.1, off_time=0.1)

        # sound buzzer
        self.buzzer.beep(on_time=0.1, off_time=0.05, n=2)

        # control servomotor
        self.servo.min()
        sleep(0.5)
        self.servo.mid()
        sleep(0.5)
        self.servo.detach()  # No signal sent

        # Red LED on
        self.led_red.on()

        # set lock status
        self.locked = True

    def unlock(self) -> None:
        """Excute unlocking sequence.

        The detail of the sequence is as follows:

        1. Red LED off
        2. Green LED Blinking
        3. Buzzer beep (3 times)
        4. Servomotor moving
        5. Green LED on
        """
        self.logger.debug("unlocking sequence started")

        # Red LED off
        self.led_red.off()

        # Green LED blinking
        self.led_green.blink(on_time=0.1, off_time=0.1)

        # sound buzzer
        self.buzzer.beep(on_time=0.1, off_time=0.05, n=3)

        # control servomotor
        self.servo.max()
        sleep(0.5)
        self.servo.mid()
        sleep(0.5)
        self.servo.detach()  # No signal sent

        # Green LED on
        self.led_green.on()

        # set lock status
        self.locked = False

    def _check_pin_overlap(self, maps: dict[str, int]):
        """Check if pin assignment is overlaped."""
        # pin numbers
        pin_numbers = [num for num in maps.values()]

        for pin in pin_numbers:
            if pin_numbers.count(pin) > 1:
                raise Exception("detect overlaped pin assignment!")
